@@ -3,8 +3,11 @@ import subprocess
 import pandas as pd
 import colorama
 import string
+import shutil
+import textwrap
 
 from .Utils import *
+from .Printers import *
 
 
 class MatrixTestRunner:
@@ -60,13 +63,18 @@ class MatrixTestRunner:
         fields_cmd.sort()
         return fields_matrix == fields_cmd
 
-    def __init__(self, cmd: str, matrix: Dict[str, List[str]], parser: Callable[[str], Any] = None):
+    def __init__(self, cmd: str, matrix: Dict[str, List[str]], parser: Callable[[str], Any] = None, enable_echo: bool = False):
+        """
+        Instantiate ``MatrixTestRunner``, checking the user input and initializing options.
+
+        For the tutorial please move to `Github <https://github.com/DavyVan/MatrixTest>`_
+
+        :param cmd: The command line template string, which follows the ``string.Formatter`` style.
+        :param matrix: The possible arguments.
+        :param parser: The parser function.
+        :param enable_echo: If ``True``, the ``stdout`` and ``stderr`` of the the command will be piped and print in real time.
         """
 
-        :param cmd:
-        :param matrix:
-        :param parser:
-        """
         colorama.init()  # enable ANSI support on Windows for colored output
 
         self.__cmd = cmd
@@ -102,42 +110,50 @@ class MatrixTestRunner:
         self.__last_result = None   # type: Optional[pd.DataFrame]
         self.__last_repeat = 0
         self.__last_aggregated_columns = []
+        # log to file
+        self.__log_enabled = False
+        self.__log_fd = None        # type: Optional[io.TextIOBase]
+        self.__log_file = None       # type: Optional[str]
+
+        # options
+        self.__option_echo = enable_echo
+        self.__terminal_width, _ = shutil.get_terminal_size()       # this is used by self.__option_echo
+
+    def run(self, repeat: int = 1, logfile: str = None) -> None:
+        """
+        Run the test suite and record the results.
+
+        :param repeat: The number of times the test should be repeated for.
+        :param logfile: If given a file path, all the output (stdout) **of this function** will also be written to that file (in append mode).
+            The content of this file will also be controlled by whether the echo feature is enabled or not
+            (see also :func:`__init__`, :func:`enable_echo`, and :func:`disable_echo`).
+        :return: None. The results will be stored internally. You can use :func:`get_last_result` to get the last result
+            as a `pandas.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_
+        """
+        # log to file
+        if logfile is not None:
+            self.__log_file = logfile
+            self.__log_enabled = True
+            self.__log_fd = open(self.__log_file, 'a')  # in append mode
+            print("Console output will be written to " + self.__log_file)
+        else:
+            self.__log_file = None
+            self.__log_fd = None
+            self.__log_enabled = False
 
-    def run(self, repeat: int = 1) -> None:
         if repeat < 1:
-            print_error("repeat must be at least 1.")
-            print_aborted()
+            print_error("repeat must be at least 1.", self.__log_fd)
+            print_aborted(self.__log_fd)
         self.__last_repeat = repeat
         self.__last_aggregated_columns = []     # clear
+        tw = textwrap.TextWrapper(width=self.__terminal_width-1)
 
         # initialize the result matrix
         results_columns = ["cmd_full"]
         results_columns += self.__arg_keys
-        # TODO: remove data type inference
         # for i in range(repeat):
         #     results_columns.append("attempt" + str(i+1))
         results = pd.DataFrame(columns=results_columns)
-        #
-        # # infer the result type from parser type hint
-        # parser_type_hints = get_type_hints(self.__parser)
-        # if "return" in parser_type_hints:
-        #     parser_return_type = parser_type_hints["return"]
-        #     df_type_str = ""
-        #     if issubclass(parser_return_type, int):
-        #         df_type_str = "int64"
-        #     elif issubclass(parser_return_type, str):
-        #         df_type_str = "string"
-        #     elif issubclass(parser_return_type, float):
-        #         df_type_str = "float64"
-        #
-        #     if df_type_str != "":
-        #         print_info("Result data type has been inferred as " + df_type_str)
-        #         new_types = {}
-        #         for i in range(repeat):
-        #             new_types["attempt" + str(i + 1)] = df_type_str
-        #         results.astype(new_types)
-        # else:
-        #     print_info("Result data type will be inferred automatically by pandas")
 
         # configure the arg fields and run
         # init
@@ -152,24 +168,44 @@ class MatrixTestRunner:
                 args[self.__arg_keys[i]] = self.__matrix[self.__arg_keys[i]][key_index[i]]
 
             current_cmd = self.__cmd.format_map(args)
-            print("Running: " + current_cmd)
+            print_plain("Running: " + current_cmd, self.__log_fd)
 
             # run
             record = {"cmd_full": current_cmd}
             record.update(args)
             for attempt in range(repeat):
-                print("Attempt " + str(attempt + 1) + "...", end="")
-                current_result = subprocess.run(current_cmd, stdout=subprocess.PIPE, text=True, shell=True)
-                if current_result.returncode != 0:
-                    print_error("Return code is " + str(current_result.returncode))
-                    print_error("STDERR:" + str(current_result.stderr))
-                    print_error("STDOUT:" + str(current_result.stdout))
+                prefix = "Attempt " + str(attempt + 1) + "..."
+                tw.initial_indent = ' '*len(prefix) + '| '
+                tw.subsequent_indent = ' '*len(prefix) + '| '
+                if self.__option_echo:
+                    print_plain(("{:-<%d}" % (self.__terminal_width-1)).format(prefix+'+'), self.__log_fd)
                 else:
-                    print_ok("Success")
-                # print(current_result.stdout)
+                    print_plain(prefix, self.__log_fd, end="")
+
+                stdout = ""
+
+                proc = subprocess.Popen(current_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, bufsize=1)
+
+                for line in proc.stdout:
+                    if self.__option_echo:
+                        print_plain(tw.fill(line), self.__log_fd)
+                    stdout += line
+
+                returncode = proc.wait()
+
+                if self.__option_echo:
+                    print_plain(("{:-<%d}" % (self.__terminal_width-1)).format(' '*len(prefix)+'+'), self.__log_fd)
+
+                if returncode != 0:
+                    print_error("Return code is " + str(returncode), self.__log_fd)
+                    print_error("STDOUT:" + str(stdout), self.__log_fd)
+                else:
+                    if self.__option_echo:
+                        print_plain(prefix, self.__log_fd, end="")
+                    print_ok("Success", self.__log_fd)
 
                 # parse the result
-                current_result_parsed = self.__parser(str(current_result.stdout))
+                current_result_parsed = self.__parser(str(stdout))
 
                 # record the result
                 if isinstance(current_result_parsed, dict):  # if the parser function returns a dict (multiple results)
@@ -192,12 +228,14 @@ class MatrixTestRunner:
                 break
 
         self.__last_result = results
-        print("All done.")
+        print_plain("All done.", self.__log_fd)
+        if self.__log_enabled:
+            self.__log_fd.close()
 
     def get_last_result(self) -> pd.DataFrame:
         """
 
-        :return: pandas.DataFrame. Return the internal dataframe storing the final results.
+        :return: Return the internal dataframe storing the final results.
         """
         return self.__last_result
 
@@ -217,7 +255,7 @@ class MatrixTestRunner:
             column = []
             for col in columns:
                 if col.startswith("attempt1_"):
-                    column.append(col.removeprefix("attempt1_"))
+                    column.append(removeprefix(col, "attempt1_"))
 
         # convert the str to List[str]. Will use List anyway in the following
         if isinstance(column, str):
@@ -290,3 +328,21 @@ class MatrixTestRunner:
                 columns_list.remove(item)
         self.__last_result.to_excel(path, sheet_name="MatrixTest", columns=columns_list, index=False)
         print_ok("Done.")
+
+    def enable_echo(self) -> None:
+        """
+        Enable echo feature, output the ``cmd``'s ``stdout`` in real time.
+
+        :return: None
+        """
+        self.__option_echo = True
+
+    def disable_echo(self) -> None:
+        """
+        Disable echo feature.
+
+        :return: None
+
+        See also: :func:`enable_echo`
+        """
+        self.__option_echo = False
